/* 
 * smview.c --
 *
 *	This file implements canvas viewport widgets.
 *
 * Portion of this code based on tkRectOval.c
 *
 * Copyright (c) 1991-1994 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 *
 * See the file "license.terms.TclTk" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "copyright.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <tk.h>
#include <sys/types.h>
#include <X11/Xutil.h>
#include <tkCanvas.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include "smInt.h"


/*
 * Prototypes for procedures defined externally
 */

void		SmRemovePortFromCell _ANSI_ARGS_ ((Cell *cell, PortItem *port));

int		SmViewportChanged _ANSI_ARGS_ ((PortItem *port));


/*
 * Prototypes for procedures defined in this file that are exported:
 */

int		SmNormalizeVector3D _ANSI_ARGS_ ((DOUBLE vin[], DOUBLE vout[], double epsilon));

void		SmFastNormalizeVector3D _ANSI_ARGS_ ((DOUBLE vin[], DOUBLE vout[]));

void		SmCrossVector3D _ANSI_ARGS_ ((DOUBLE a[], DOUBLE b[], DOUBLE c[]));

void		SmDisplayViewport _ANSI_ARGS_ ((Tk_Canvas canvas, Tk_Item *itemPtr, Display *dpy, Drawable dst,
						int x, int y, int width, int height));

void		SmDisplayObjects _ANSI_ARGS_ ((PortItem *port, int pick));

void		SmSelectViewport _ANSI_ARGS_ ((PortItem *port));

XVisualInfo	*SmGetVisualInfo _ANSI_ARGS_ ((Display *dpy, Tk_Window tkwin, int viewport));

int		SmConfigureViewport _ANSI_ARGS_ ((Tcl_Interp *interp,
						  Tk_Canvas canvas, Tk_Item *itemPtr, int argc, char **argv, int flags));

void		SmCreateDrawable _ANSI_ARGS_ ((PortItem *port));


/*
 * Prototypes for procedures defined in this file that are static:
 */

static int	CreateViewport _ANSI_ARGS_ ((Tcl_Interp *interp,
					     Tk_Canvas canvas, struct Tk_Item *itemPtr, int argc, char **argv));

static int	PortCoords _ANSI_ARGS_ ((Tcl_Interp *interp, Tk_Canvas canvas, Tk_Item *itemPtr,
					 int argc, char **argv));

static void	DeleteViewport _ANSI_ARGS_ ((Tk_Canvas canvas, Tk_Item *itemPtr, Display *dpy));

static void	ComputeViewportBbox _ANSI_ARGS_ ((Tk_Canvas canvas, PortItem *port));

static void	SetupLightSources _ANSI_ARGS_ ((PortItem *port));

static double	PortToPoint _ANSI_ARGS_ ((Tk_Canvas canvas, Tk_Item *itemPtr, double *pointPtr));

static int	PortToArea _ANSI_ARGS_ ((Tk_Canvas canvas, Tk_Item *itemPtr, double *areaPtr));

static void	ScaleViewport _ANSI_ARGS_ ((Tk_Canvas canvas, Tk_Item *itemPtr, double originX, double originY,
					    double scaleX, double scaleY));

static void	TranslateViewport _ANSI_ARGS_ ((Tk_Canvas canvas, Tk_Item *itemPtr, double deltaX, double deltaY));

static int	ComputeViewTransform _ANSI_ARGS_ ((PortItem *port));

static int	Equal _ANSI_ARGS_ ((DOUBLE a, DOUBLE b, double epsilon));

static int	EqualVector3D _ANSI_ARGS_ ((DOUBLE a[], DOUBLE b[], double epsilon));

static int	DrawObject _ANSI_ARGS_ ((PortItem *port, Object *obj, GLuint name));


/*
 * Data structure declarations
 */

extern Cell *cell;
extern Tk_ConfigSpec canvasConfigSpecs[];
extern Tk_ConfigSpec widgetConfigSpecs[];


/*
 * The structures below defines the viewport item type
 * by means of procedures that can be invoked by generic item code.
 */

Tk_ItemType tkViewportType = {
    "viewport",				/* name */
    sizeof(PortItem),			/* itemSize */
    CreateViewport,			/* createProc */
    canvasConfigSpecs,			/* configSpecs */
    SmConfigureViewport,		/* configureProc */
    PortCoords,				/* coordProc */
    DeleteViewport,			/* deleteProc */
    SmDisplayViewport,			/* displayProc */
    0,					/* alwaysRedraw */
    PortToPoint,			/* pointProc */
    PortToArea,				/* areaProc */
    (Tk_ItemPostscriptProc *) NULL,	/* postscriptProc */
    ScaleViewport,			/* scaleProc */
    TranslateViewport,			/* translateProc */
    (Tk_ItemIndexProc *) NULL,		/* indexProc */
    (Tk_ItemCursorProc *) NULL,		/* icursorProc */
    (Tk_ItemSelectionProc *) NULL,	/* selectionProc */
    (Tk_ItemInsertProc *) NULL,		/* insertProc */
    (Tk_ItemDCharsProc *) NULL,		/* dTextProc */
    (Tk_ItemType *) NULL		/* nextPtr */
};



/*
 *--------------------------------------------------------------
 *
 * CreateViewport:
 *
 * create a new viewport item in a canvas.
 *
 *--------------------------------------------------------------
 */

static int
CreateViewport(interp, canvas, itemPtr, argc, argv)
    Tcl_Interp *interp;
    Tk_Canvas canvas;
    Tk_Item *itemPtr;
    int argc;
    char **argv;
{
    int i;
    XGCValues gcValues;
    unsigned long mask;
    PortItem *port = (PortItem *) itemPtr;

    if (argc < 4) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		Tk_PathName(Tk_CanvasTkwin(canvas)), "\" create ",
		itemPtr->typePtr->name, " x1 y1 x2 y2 ?options?",
		(char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Carry out initialization that is needed in order to clean
     * up after errors during the the remainder of this procedure.
     */

    port->redraw = 1;
    port->dpy = Tk_Display(Tk_CanvasTkwin(canvas));
    port->canvas = canvas;
    port->tkwin = NULL;
    port->widgetCmd = NULL;
    mask = GCFunction | GCPlaneMask;
    gcValues.function = GXcopy;
    gcValues.plane_mask = AllPlanes;
    port->initialized = 0;
    port->visinfo = NULL;
    port->pm = None;
    port->pmGC = Tk_GetGC(Tk_CanvasTkwin(canvas), mask, &gcValues);
    port->glxpm = None;
    port->glxpmGC = None;
    port->outlineColor = NULL;
    port->fillColor = NULL;
    port->outlineGC = None;
    port->interp = interp;
    port->invalid = 0;
    port->cell = NULL;
    port->next = NULL;
    port->pick[0] = 0;
    port->ambientLight = NULL;
    port->ambient[0] = port->ambient[1] = port->ambient[2] = 0;
    port->ambient[3] = 1.0;
    port->depthcue.color = NULL;
    port->hsr = 1;
    port->epsilon = 1e-10;
    port->srcx = port->srcy = 0;
    port->srcw = port->srch = 0;
    port->invertedVisibleMask = 0;
    port->visibleMaskStr = NULL;

    for (i = 0; i < 8; i++) {
	port->lights[i].on = 0;
	port->lights[i].ambientColor = port->lights[i].diffuseColor = port->lights[i].specularColor = NULL;
    }

    Tcl_InitHashTable(&port->visibleHash, TCL_STRING_KEYS);

    /*
     * Process the arguments to fill in the item record.
     */

    if ((Tk_CanvasGetCoord(interp, canvas, argv[0], &port->bbox[0]) != TCL_OK)
	|| (Tk_CanvasGetCoord(interp, canvas, argv[1], &port->bbox[1]) != TCL_OK)
	|| (Tk_CanvasGetCoord(interp, canvas, argv[2], &port->bbox[2]) != TCL_OK)
	|| (Tk_CanvasGetCoord(interp, canvas, argv[3], &port->bbox[3]) != TCL_OK)) {
	DeleteViewport(canvas, itemPtr, port->dpy);
	return TCL_ERROR;
    }
    port->w = ceil(port->bbox[2] - port->bbox[0]);
    port->h = ceil(port->bbox[3] - port->bbox[1]);

    /* get desired visual & context */

    if (!(port->visinfo = SmGetVisualInfo(port->dpy, Tk_CanvasTkwin(canvas), 0))) {
	Tcl_AppendResult(interp, "failed to find a valid visual", (char *) NULL);
	DeleteViewport(canvas, itemPtr, port->dpy);
	return TCL_ERROR;
    }

    /* create drawing buffer */

    SmCreateDrawable(port);
    if (SmConfigureViewport(interp, canvas, itemPtr, argc-4, argv+4, 0) != TCL_OK) {
	DeleteViewport(canvas, itemPtr, port->dpy);
	return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * PortCoords:
 *
 * process the "coords" widget command on a viewport.
 *
 *--------------------------------------------------------------
 */

static int
PortCoords(interp, canvas, itemPtr, argc, argv)
    Tcl_Interp *interp;
    Tk_Canvas canvas;
    Tk_Item *itemPtr;
    int argc;
    char **argv;
{
    PortItem *port = (PortItem *) itemPtr;
    double x1, y1, x2, y2;
    char c0[TCL_DOUBLE_SPACE], c1[TCL_DOUBLE_SPACE];
    char c2[TCL_DOUBLE_SPACE], c3[TCL_DOUBLE_SPACE];

    if (argc == 0) {
	Tcl_PrintDouble(interp, (double) port->bbox[0], c0);
	Tcl_PrintDouble(interp, (double) port->bbox[1], c1);
	Tcl_PrintDouble(interp, (double) port->bbox[2], c2);
	Tcl_PrintDouble(interp, (double) port->bbox[3], c3);
	Tcl_AppendResult(interp, c0, " ", c1, " ", c2, " ", c3,	(char *) NULL);
    } else if (argc == 4) {
	if ((Tk_CanvasGetCoord(interp, canvas, argv[0], &x1) != TCL_OK)
	    || (Tk_CanvasGetCoord(interp, canvas, argv[1], &y1) != TCL_OK)
	    || (Tk_CanvasGetCoord(interp, canvas, argv[2], &x2) != TCL_OK)
	    || (Tk_CanvasGetCoord(interp, canvas, argv[3], &y2) != TCL_OK)) {
	    return TCL_ERROR;
	}

	/* make sure viewport isn't too big */

	port->w = x2 - x1;
	port->h = y2 - y1;
	port->bbox[0] = x1;
	port->bbox[1] = y1;
	port->bbox[2] = x2;
	port->bbox[3] = y2;
	ComputeViewportBbox(canvas, port);

	SmCreateDrawable(port);
	ComputeViewTransform(port);
	port->redraw = 1;
    } else {
	sprintf(interp->result,	"wrong # coordinates:  expected 0 or 4, got %d", argc);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SmConfigureViewport:
 *
 * configure a viewport item or a viewport widget.
 *
 *--------------------------------------------------------------
 */

int
SmConfigureViewport(interp, canvas, itemPtr, argc, argv, flags)
     Tcl_Interp *interp;
     Tk_Canvas canvas;
     Tk_Item *itemPtr;
     int argc;
     char **argv;
     int flags;
{
    GC newGC;
    unsigned long mask;
    XGCValues gcValues;
    Tk_Window tkwin;
    Tk_ConfigSpec *spec;
    PortItem *port = (PortItem *) itemPtr;
    XColor *ambient = port->ambientLight;

    if (port->canvas) {
	tkwin = Tk_CanvasTkwin(port->canvas);
	spec = canvasConfigSpecs;
    } else {
	tkwin = (Tk_Window) canvas;
	spec = widgetConfigSpecs;
    }

    if (Tk_ConfigureWidget(interp, tkwin, spec, argc, argv, (char *) port, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * A few of the options require additional processing, such as
     * graphics contexts.
     */
    if (port->width < 1) {
	port->width = 1;
    }
    if (port->outlineColor == NULL) {
	newGC = None;
    } else {
	gcValues.foreground = port->outlineColor->pixel;
	gcValues.cap_style = CapProjecting;
	gcValues.line_width = port->width;
	mask = GCForeground|GCCapStyle|GCLineWidth;
	newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (port->outlineGC != None) {
	Tk_FreeGC(Tk_Display(tkwin), port->outlineGC);
    }
    port->outlineGC = newGC;
    port->redraw = 1;

    if (port->canvas) {
	ComputeViewportBbox(canvas, port);
    }

    if (port->ambientLight != ambient) {
	port->ambient[0] = port->ambientLight->red / 65535.0;
	port->ambient[1] = port->ambientLight->green / 65535.0;
	port->ambient[2] = port->ambientLight->blue / 65535.0;
    }
    if ((!port->initialized) || SmViewportChanged(port)) {
	if ((port->invalid = ComputeViewTransform(port)))
	  return TCL_ERROR;
	port->initialized = 1;
    }

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DeleteViewport:
 *
 * clean up the data structure associated with a viewport item.
 *
 *--------------------------------------------------------------
 */

static void
DeleteViewport(canvas, itemPtr, dpy)
    Tk_Canvas canvas;
    Tk_Item *itemPtr;
    Display *dpy;
{
    int i;
    PortItem *port = (PortItem *) itemPtr;

    Tcl_DeleteHashTable(&port->visibleHash);

    for (i = 0; i < 8; i++) {
	if (port->lights[i].on) {
	    Tk_FreeColor(port->lights[i].ambientColor);
	    Tk_FreeColor(port->lights[i].diffuseColor);
	    Tk_FreeColor(port->lights[i].specularColor);
	}
    }
    if (port->ambientLight != NULL) {
	Tk_FreeColor(port->ambientLight);
    }
    if (port->outlineColor != NULL) {
	Tk_FreeColor(port->outlineColor);
    }
    if (port->fillColor != NULL) {
	Tk_FreeColor(port->fillColor);
    }
    if (port->depthcue.color != NULL) {
	Tk_FreeColor(port->depthcue.color);
    }
    if (port->outlineGC != None) {
	Tk_FreeGC(dpy, port->outlineGC);
    }
    if (port->cell != NULL) {
	SmRemovePortFromCell(port->cell, port);
    }
    if (port->glxpm != None) {
	glXDestroyGLXPixmap(port->dpy, port->glxpm);
    }
    if (port->glxpmGC != None) {
	glXDestroyContext(port->dpy, port->glxpmGC);
    }
    if (port->pm != None) {
	XFreePixmap(port->dpy, port->pm);
    }
    if (port->pmGC != None) {
	Tk_FreeGC(dpy, port->pmGC);
    }
}

/*
 *--------------------------------------------------------------
 *
 * ComputeViewportBbox:
 *
 * compute the bounding box of all the pixels that may be drawn
 * as part of a viewport.
 *
 *--------------------------------------------------------------
 */

/* ARGSUSED */
static void
ComputeViewportBbox(canvas, port)
    Tk_Canvas canvas;
    PortItem *port;
{
    int bloat;

    /*
     * Make sure that the first coordinates are the lowest ones.
     */

    if (port->bbox[1] > port->bbox[3]) {
	double tmp;
	tmp = port->bbox[3];
	port->bbox[3] = port->bbox[1];
	port->bbox[1] = tmp;
    }
    if (port->bbox[0] > port->bbox[2]) {
	double tmp;
	tmp = port->bbox[2];
	port->bbox[2] = port->bbox[0];
	port->bbox[0] = tmp;
    }

    if (port->outlineColor == NULL) {
	bloat = 0;
    } else {
	bloat = (port->width+1)/2;
    }
    port->header.x1 = port->bbox[0] - bloat;
    port->header.y1 = port->bbox[1] - bloat;
    port->header.x2 = port->bbox[2] + bloat + 1;
    port->header.y2 = port->bbox[3] + bloat + 1;
}

/*
 *--------------------------------------------------------------
 *
 * SmGetVisualInfo:
 *
 * return visualinfo for a window.
 *
 *--------------------------------------------------------------
 */

XVisualInfo *
SmGetVisualInfo(dpy,  tkwin, viewport)
     Display *dpy;
     Tk_Window tkwin;
     int viewport;
{
    XWindowAttributes attr;
    XVisualInfo vinfo, *visinfo;
    long vinfoMask;
    int nitems;
    int attribCanvas[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 2,
	GLX_GREEN_SIZE, 2,
	GLX_BLUE_SIZE, 2,
	None
    };
    int attribViewport[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 2,
	GLX_GREEN_SIZE, 2,
	GLX_BLUE_SIZE, 2,
	GLX_DOUBLEBUFFER,
	None
    };

    Tk_MapWindow(tkwin);
    vinfoMask = VisualIDMask;
    XGetWindowAttributes(dpy, Tk_WindowId(tkwin), &attr);
    vinfo.visualid = XVisualIDFromVisual(attr.visual);
    visinfo = XGetVisualInfo(dpy, vinfoMask, &vinfo, &nitems);
    if (viewport) {
/*
	attribViewport[2] = attribViewport[4] = attribViewport[6] = visinfo->depth / 3;
*/
	return glXChooseVisual(dpy, visinfo->screen, attribViewport);
    } else {
	attribCanvas[2] = attribCanvas[4] = attribCanvas[6] = visinfo->depth / 3;
	return glXChooseVisual(dpy, visinfo->screen, attribCanvas);
    }
}

/*
 *--------------------------------------------------------------
 *
 * SmDisplayViewport:
 *
 * draw a viewport item in a given drawable.
 *
 *--------------------------------------------------------------
 */

void
SmDisplayViewport(canvas, itemPtr, dpy, drawable, x, y, width, height)
     Tk_Canvas canvas;
     Tk_Item *itemPtr;
     Display *dpy;
     Drawable drawable;
     int x, y, width, height;
{
    short x1, y1, x2, y2;
    int w, h;
    int srcx, srcy, tx, ty;
    int srcw, srch;
    int tgtx, tgty;
    int immediate, extra;
    float rgba[4];
    unsigned int twidth, theight, tborder, tdepth;
    Window troot;
    GLbitfield clear_mask;

    PortItem *port = (PortItem *) itemPtr;
    Cell *cell = port->cell;
    Tk_Window tkwin;

    if ((port->pm == None) && (port->tkwin == NULL)) return;

    if (canvas) {
	tkwin = Tk_CanvasTkwin(canvas);
	immediate = (drawable == Tk_WindowId(tkwin));

	/*
	 * Compute the screen coordinates of the bounding box for the item.
	 * Make sure that the bbox is at least one pixel large, since some
	 * X servers will die if it isn't.
	 */

	if (immediate) {
	    Tk_CanvasWindowCoords(canvas, port->bbox[0], port->bbox[1], &x1, &y1);
	    Tk_CanvasWindowCoords(canvas, port->bbox[2], port->bbox[3], &x2, &y2);
	}
	else {
	    Tk_CanvasDrawableCoords(canvas, port->bbox[0], port->bbox[1], &x1, &y1);
	    Tk_CanvasDrawableCoords(canvas, port->bbox[2], port->bbox[3], &x2, &y2);
	}

	w = x2 - x1;
	h = y2 - y1;
	if (w <= 0) w = 1;
	if (h <= 0) h = 1;
    
	/*
	 * Display filled part first (if wanted), then port objects, then
	 * outline (if wanted).
	 */

	XGetGeometry(dpy, drawable, &troot, &tx, &ty, &twidth, &theight, &tborder, &tdepth);

	if (immediate) {
	    
	    /* need to get canvas's border width and highlight thickness */

	    extra = ((TkCanvas *) canvas)->borderWidth + ((TkCanvas *) canvas)->highlightWidth;
	    twidth -= (extra + extra);
	    theight -= (extra + extra);
	    if (x1 < extra) {
		srcx = extra;
		srcw = w + x1 - extra;
	    }
	    else {
		srcx = x1;
		srcw = w;
	    }
	    if (y1 < extra) {
		srcy = extra;
		srch = h + y1 - extra;
	    }
	    else {
		srcy = y1;
		srch = h;
	    }
	    if (srcx + srcw > twidth) {
		srcw = twidth - srcx;
	    }
	    if (srcy + srch > theight) {
		srch = theight - srcy;
	    }
	}
	else {
	    if (x1 < tx) {
		srcx = tx;
		srcw = w - (tx - x1);
	    }
	    else {
		srcx = x1;
		srcw = w;
	    }
	    if (y1 < ty) {
		srcy = ty;
		srch = h - (ty - y1);
	    }
	    else {
		srcy = y1;
		srch = h;
	    }
	    if (srcx + srcw > tx + twidth) {
		srcw = tx + twidth - srcx;
	    }
	    if (srcy + srch > ty + theight) {
		srch = ty + theight - srcy;
	    }
	}
	if ((srcw < 1) || (srch < 1)) return;
	tgtx = (srcx - x1);
	tgty = (srcy - y1);

	/* force redraw if previous drawing area does not full contain current drawing area */
	if ((!port->redraw) &&
	    ((tgtx < port->srcx) ||
	     (tgty < port->srcy) ||
	     (tgtx + srcw > port->srcx + port->srcw) ||
	     (tgty + srch > port->srcy + port->srch))) {
	    port->redraw = 1;

#ifdef DEBUG_VIEW
	printf("current drawing area: (%d %d %u %u), previous drawing area: (%d %d %u %u)\n",
	       tgtx, tgty, srcw, srch,
	       port->srcx, port->srcy, port->srcw, port->srch);
#endif
	}
    } else {
	w = width - port->width;
	h = height - port->width;
	x1 = y1 = port->width / 2;
    }

    if (port->redraw) {
	    
	/* draw objects */

	SmSelectViewport(port);

	/* set up color buffer clearing value */
	glClearColor((GLclampf) port->fillColor->red / 65535.0,
		     (GLclampf) port->fillColor->green / 65535.0,
		     (GLclampf) port->fillColor->blue / 65535.0,
		     0.0);

	/* clear buffers */
	clear_mask = GL_COLOR_BUFFER_BIT;

	if (port->hsr) {
	    glEnable(GL_DEPTH_TEST);
	    clear_mask |= GL_DEPTH_BUFFER_BIT;
	} else {
	    glDisable(GL_DEPTH_TEST);
	}
	glClear(clear_mask);

	/* enable lighting */
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, port->ambient);
	SetupLightSources(port);

	if (port->depthcue.color) {

	    rgba[0] = port->depthcue.color->red / 65535;
	    rgba[1] = port->depthcue.color->red / 65535;
	    rgba[2] = port->depthcue.color->red / 65535;
	    rgba[3] = 1.0;
	    
	    glFogi(GL_FOG_MODE, port->depthcue.mode);
	    glFogfv(GL_FOG_COLOR, rgba);
	    if ((port->depthcue.mode == GL_EXP) || (port->depthcue.mode == GL_EXP2)) {
		glFogf(GL_FOG_DENSITY, (float) port->depthcue.density);
	    } else if (port->depthcue.mode == GL_LINEAR) {
		glFogf(GL_FOG_START, (float) port->depthcue.start);
		glFogf(GL_FOG_END, (float) port->depthcue.end);
	    }
	    glEnable(GL_FOG);
	} else {
	    glDisable(GL_FOG);
	}

	if ((!port->invalid) && (cell != NULL)) {
	    SmDisplayObjects(port, 0);
	}
	if (canvas) {
	    port->srcx = tgtx;
	    port->srcy = tgty;
	    port->srcw = srcw;
	    port->srch = srch;
	    port->redraw = 0;

#ifdef DEBUG_VIEW
	    printf("drawing area: %d %d %u %u, destination origin: %d %d\n", tgtx, tgty, srcw, srch, srcx, srcy);
#endif
	} else {
	    glXSwapBuffers(dpy, drawable);
	}
	glXWaitGL();
    }

    if (canvas) {
	if ((port->pmGC != None) && (port->pm != None)) {
	    XCopyArea(dpy, port->pm, drawable, port->pmGC, tgtx, tgty,
		      srcw, srch, srcx, srcy);
	}
    }

    if (port->outlineGC != None) {
	XDrawRectangle(dpy, drawable, port->outlineGC, x1, y1, w, h);
    }
}

/*
 *--------------------------------------------------------------
 *
 * SetupLightSources:
 *
 * set up all light sources.
 *
 *--------------------------------------------------------------
 */
static void
SetupLightSources(port)
     PortItem *port;
{
    int i;
    LightSource *l;
    float ambientRGBA[4], diffuseRGBA[4], specularRGBA[4];
    static GLenum source[] = {
	GL_LIGHT0,
	GL_LIGHT1,
	GL_LIGHT2,
	GL_LIGHT3,
	GL_LIGHT4,
	GL_LIGHT5,
	GL_LIGHT6,
	GL_LIGHT7
    };

    for (i = 0; i < 8; i++) {
	l = &port->lights[i];

	if (l->on) {

	    glEnable(source[i]);

	    /* convert XColor to RGBA */
	    ambientRGBA[0] = l->ambientColor->red / 65535;
	    ambientRGBA[1] = l->ambientColor->green / 65535;
	    ambientRGBA[2] = l->ambientColor->blue / 65535;
	    ambientRGBA[3] = 1.0;

	    diffuseRGBA[0] = l->diffuseColor->red / 65535;
	    diffuseRGBA[1] = l->diffuseColor->green / 65535;
	    diffuseRGBA[2] = l->diffuseColor->blue / 65535;
	    diffuseRGBA[3] = 1.0;

	    specularRGBA[0] = l->specularColor->red / 65535;
	    specularRGBA[1] = l->specularColor->green / 65535;
	    specularRGBA[2] = l->specularColor->blue / 65535;
	    specularRGBA[3] = 1.0;

	    /* create light source */
	    glLightfv(source[i], GL_AMBIENT, ambientRGBA);
	    glLightfv(source[i], GL_DIFFUSE, diffuseRGBA);
	    glLightfv(source[i], GL_SPECULAR, specularRGBA);
	    glLightfv(source[i], GL_POSITION, l->position);
	    glLightfv(source[i], GL_SPOT_DIRECTION, l->ndirection);
	    glLightf(source[i], GL_SPOT_EXPONENT, l->exponent);
	    glLightf(source[i], GL_SPOT_CUTOFF, l->cutoff);
	    glLightf(source[i], GL_CONSTANT_ATTENUATION, l->attenuation[0]);
	    glLightf(source[i], GL_LINEAR_ATTENUATION, l->attenuation[1]);
	    glLightf(source[i], GL_QUADRATIC_ATTENUATION, l->attenuation[2]);
	} else {
	    glDisable(source[i]);
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * PortToPoint:
 *
 * computes the distance from a given point to a given viewport,
 * in canvas units.
 *
 *--------------------------------------------------------------
 */

/* ARGSUSED */
    static double
PortToPoint(canvas, itemPtr, pointPtr)
    Tk_Canvas canvas;
    Tk_Item *itemPtr;
    double *pointPtr;
{
    PortItem *port = (PortItem *) itemPtr;
    double xDiff, yDiff, x1, y1, x2, y2, inc, tmp;

    /*
     * Generate a new larger viewport that includes the border width, if there is one.
     */

    x1 = port->bbox[0];
    y1 = port->bbox[1];
    x2 = port->bbox[2];
    y2 = port->bbox[3];
    if (port->outlineGC != None) {
	inc = port->width/2.0;
	x1 -= inc;
	y1 -= inc;
	x2 += inc;
	y2 += inc;
    }

    /*
     * If the point is inside the viewport, handle specially:
     * distance is 0 if viewport is filled, otherwise compute
     * distance to nearest edge of viewport and subtract width
     * of edge.
     */

    if ((pointPtr[0] >= x1) && (pointPtr[0] < x2) && (pointPtr[1] >= y1) && (pointPtr[1] < y2)) {
	if ((port->fillColor != NULL) || (port->outlineGC == None)) {
	    return 0.0;
	}
	xDiff = pointPtr[0] - x1;
	tmp = x2 - pointPtr[0];
	if (tmp < xDiff) {
	    xDiff = tmp;
	}
	yDiff = pointPtr[1] - y1;
	tmp = y2 - pointPtr[1];
	if (tmp < yDiff) {
	    yDiff = tmp;
	}
	if (yDiff < xDiff) {
	    xDiff = yDiff;
	}
	xDiff -= port->width;
	if (xDiff < 0.0) {
	    return 0.0;
	}
	return xDiff;
    }

    /*
     * Point is outside viewport.
     */

    if (pointPtr[0] < x1) {
	xDiff = x1 - pointPtr[0];
    } else if (pointPtr[0] > x2)  {
	xDiff = pointPtr[0] - x2;
    } else {
	xDiff = 0;
    }

    if (pointPtr[1] < y1) {
	yDiff = y1 - pointPtr[1];
    } else if (pointPtr[1] > y2)  {
	yDiff = pointPtr[1] - y2;
    } else {
	yDiff = 0;
    }

    return hypot(xDiff, yDiff);
}

/*
 *--------------------------------------------------------------
 *
 * PortToArea:
 *
 * determine whether an item lies entirely inside, entirely
 * outside, or overlapping a given viewport.
 *
 *--------------------------------------------------------------
 */

/* ARGSUSED */
static int
PortToArea(canvas, itemPtr, areaPtr)
    Tk_Canvas canvas;
    Tk_Item *itemPtr;
    double *areaPtr;
{
    PortItem *port = (PortItem *) itemPtr;
    double halfWidth;

    halfWidth = port->width/2.0;
    if (port->outlineGC == None) {
	halfWidth = 0.0;
    }

    if ((areaPtr[2] <= (port->bbox[0] - halfWidth))
	|| (areaPtr[0] >= (port->bbox[2] + halfWidth))
	|| (areaPtr[3] <= (port->bbox[1] - halfWidth))
	|| (areaPtr[1] >= (port->bbox[3] + halfWidth))) {
	return -1;
    }
    if ((port->fillColor == NULL) && (port->outlineGC != None)
	&& (areaPtr[0] >= (port->bbox[0] + halfWidth))
	&& (areaPtr[1] >= (port->bbox[1] + halfWidth))
	&& (areaPtr[2] <= (port->bbox[2] - halfWidth))
	&& (areaPtr[3] <= (port->bbox[3] - halfWidth))) {
	return -1;
    }
    if ((areaPtr[0] <= (port->bbox[0] - halfWidth))
	&& (areaPtr[1] <= (port->bbox[1] - halfWidth))
	&& (areaPtr[2] >= (port->bbox[2] + halfWidth))
	&& (areaPtr[3] >= (port->bbox[3] + halfWidth))) {
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * ScaleViewport:
 *
 * rescale a viewport item.
 *
 *--------------------------------------------------------------
 */

static void
ScaleViewport(canvas, itemPtr, originX, originY, scaleX, scaleY)
    Tk_Canvas canvas;
    Tk_Item *itemPtr;
    double originX, originY;
    double scaleX;
    double scaleY;
{
    PortItem *port = (PortItem *) itemPtr;

    port->bbox[0] = originX + scaleX*(port->bbox[0] - originX);
    port->bbox[1] = originY + scaleY*(port->bbox[1] - originY);
    port->bbox[2] = originX + scaleX*(port->bbox[2] - originX);
    port->bbox[3] = originY + scaleY*(port->bbox[3] - originY);
    ComputeViewportBbox(canvas, port);

    SmCreateDrawable(port);
    ComputeViewTransform(port);
    port->redraw = 1;
}

/*
 *--------------------------------------------------------------
 *
 * TranslateViewport:
 *
 * move a viewport by a given amount.
 *
 *--------------------------------------------------------------
 */

static void
TranslateViewport(canvas, itemPtr, deltaX, deltaY)
    Tk_Canvas canvas;
    Tk_Item *itemPtr;
    double deltaX, deltaY;
{
    PortItem *port = (PortItem *) itemPtr;

    port->bbox[0] += deltaX;
    port->bbox[1] += deltaY;
    port->bbox[2] += deltaX;
    port->bbox[3] += deltaY;
    ComputeViewportBbox(canvas, port);
}

/*
 *--------------------------------------------------------------
 *
 * ComputeViewTransform:
 *
 * compute the view transform matrix.
 *
 *--------------------------------------------------------------
 */

static int
ComputeViewTransform(port)
     PortItem *port;
{
    Tcl_Interp *interp = port->interp;
    double epsilon = port->epsilon;

    DOUBLE vup[3];
    DOUBLE rx[3], ry[3], rz[3];
    DOUBLE *mr, msh[16] = {
	1, 0, 0, 0,
	0, 1, 0, 0,
	0, 0, 1, 0,
	0, 0, 0, 1,
    };

    if (SmNormalizeVector3D(port->vpn, rz, epsilon)) {
	Tcl_AppendResult(interp, "invalid VPN (view projection normal) value.", (char *) NULL);
	return 1;
    }

    if (SmNormalizeVector3D(port->vup, vup, epsilon)) {
	Tcl_AppendResult(interp, "invalid VUP (view up) value.", (char *) NULL);
	return 1;
    }

    SmCrossVector3D(port->vup, rz, rx);
    if (SmNormalizeVector3D(rx, rx, epsilon)) {
	if (EqualVector3D(vup, rz, epsilon)) {
	    Tcl_AppendResult(interp, "VPN and VUP must not point in the same direction.", (char *) NULL);
	} else {
	    Tcl_AppendResult(interp, "invalid VUP (view up) value.", (char *) NULL);
	}
	return 1;
    }

    SmCrossVector3D(rz, rx, ry);

    if (Equal(port->maxu, port->minu, epsilon) ||
	Equal(port->maxv, port->minv, epsilon) ||
	(port->maxu < port->minu) ||
	(port->maxv < port->minv)) {
	Tcl_AppendResult(interp, "invalid view window specification.", (char *) NULL);
	return 1;
    }

    if (port->prp[2] <= epsilon) {
	Tcl_AppendResult(interp, "prp (projection reference point) must have a positive z component.", (char *) NULL);
	return 1;
    }

    memcpy(port->msh, msh, sizeof(DOUBLE) * 16);

    /* set up rotation matrix */
    mr = port->mr;
    mr[0] = rx[0];
    mr[1] = ry[0];
    mr[2] = rz[0];
    mr[3] = 0;

    mr[4] = rx[1];
    mr[5] = ry[1];
    mr[6] = rz[1];
    mr[7] = 0;

    mr[8] = rx[2];
    mr[9] = ry[2];
    mr[10] = rz[2];
    mr[11] = 0;

    mr[12] = 0;
    mr[13] = 0;
    mr[14] = 0;
    mr[15] = 1;

    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * SmSelectViewport:
 *
 * make viewport's transform matrices current.
 *
 *--------------------------------------------------------------
 */

void
SmSelectViewport(port)
     PortItem *port;
{
    /* set up the right OpenGL/Mesa context */
    if (port->canvas) {
	glXMakeCurrent(port->dpy, port->glxpm, port->glxpmGC);
    } else {
	glXMakeCurrent(port->dpy, Tk_WindowId(port->tkwin), port->glxpmGC);
    }
    glViewport(0, 0, port->w, port->h);

    /* set up PROJECTION transform matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    switch(port->projection) {
      case SM_PORT_PARALLEL:
	port->msh[8] = -port->prp[0] / port->prp[2];
	port->msh[9] = -port->prp[1] / port->prp[2];
	glOrtho(port->minu - port->prp[0],
		port->maxu - port->prp[0],
		port->minv - port->prp[1],
		port->maxv - port->prp[1],
		port->prp[2], port->prp[2] - port->bp);
#ifdef USE_DOUBLE
	glMultMatrixd(port->msh);
#else
	glMultMatrixf(port->msh);
#endif
	break;

      case SM_PORT_PERSPECTIVE:
	glFrustum(port->minu - port->prp[0],
		  port->maxu - port->prp[0],
		  port->minv - port->prp[1],
		  port->maxv - port->prp[1],
		  port->prp[2], port->prp[2] - port->bp);
	break;
    }
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

#ifdef USE_DOUBLE
    glTranslated(-port->prp[0], -port->prp[1], -port->prp[2]);
    glMultMatrixd(port->mr);
    glTranslated(-port->vrp[0], -port->vrp[1], -port->vrp[2]);
#else
    glTranslatef(-port->prp[0], -port->prp[1], -port->prp[2]);
    glMultMatrixf(port->mr);
    glTranslatef(-port->vrp[0], -port->vrp[1], -port->vrp[2]);
#endif
}

/*
 *--------------------------------------------------------------
 *
 * SmNormalizeVector3D:
 *
 * normalize a 3D vector that is represented as three scalar
 * values.
 *
 *--------------------------------------------------------------
 */

int
SmNormalizeVector3D(vin, vout, epsilon)
     DOUBLE vin[];
     DOUBLE vout[];
     double epsilon;
{
    DOUBLE mag;

    mag = (DOUBLE) sqrt((double) (vin[0] * vin[0] + vin[1] * vin[1] + vin[2] * vin[2]));
    if (mag <= epsilon) {
	vout[0] = vout[1] = vout[2] = 0;
	return 1;
    }
    mag = 1.0 / mag;
    vout[0] = vin[0] * mag;
    vout[1] = vin[1] * mag;
    vout[2] = vin[2] * mag;
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * SmFastNormalizeVector3D:
 *
 * normalize a 3D vector that is represented as three scalar
 * values.
 *
 *--------------------------------------------------------------
 */

void
SmFastNormalizeVector3D(vin, vout)
     DOUBLE vin[];
     DOUBLE vout[];
{
    DOUBLE mag;

    mag = (DOUBLE) 1.0 / sqrt((double) (vin[0] * vin[0] + vin[1] * vin[1] + vin[2] * vin[2]));
    vout[0] = vin[0] * mag;
    vout[1] = vin[1] * mag;
    vout[2] = vin[2] * mag;
}

/*
 *--------------------------------------------------------------
 *
 * SmCrossVector3D:
 *
 * compute the cross product of two 3D vectors, each of which
 * is represented by 3 scalar values.
 *
 *--------------------------------------------------------------
 */

void
SmCrossVector3D(a, b, c)
     DOUBLE a[];
     DOUBLE b[];
     DOUBLE c[];
{
    c[0] = a[1] * b[2] - a[2] * b[1];
    c[1] = a[2] * b[0] - a[0] * b[2];
    c[2] = a[0] * b[1] - a[1] * b[0];
}

/*
 *--------------------------------------------------------------
 *
 * Equal:
 *
 * determine if two floating point values are close enough
 * (within epsilon) to one another to be considered equal.
 *
 *--------------------------------------------------------------
 */

static int
Equal(DOUBLE a, DOUBLE b, double epsilon)
{
    return (fabs((double) (a - b)) <= epsilon) ? 1 : 0;
}

/*
 *--------------------------------------------------------------
 *
 * EqualVector3D:
 *
 * determine if two 3D vectors are close enough (each scalar
 * componnet within epsilon of its counterpart) to one another
 * to be considered equal.
 *
 *--------------------------------------------------------------
 */

static int
EqualVector3D(a, b, epsilon)
     DOUBLE a[], b[];
     double epsilon;
{
    return ((fabs((double) (a[0] - b[0])) <= epsilon)
	    && (fabs((double) (a[1] - b[1])) <= epsilon)
	    && (fabs((double) (a[2] - b[2])) <= epsilon));
}

/*
 *--------------------------------------------------------------
 *
 * SmDisplayObjects:
 *
 * draw surfacea in a given drawable.
 *
 *--------------------------------------------------------------
 */

void
SmDisplayObjects(port, pick)
     PortItem *port;
     int pick;
{
    Object *obj;
    GLuint name = pick;

    if (pick) {
	name = 1;
    } else {
	name = 0;

	/* hidden surface removal */
	if (port->hsr) {
	    glEnable(GL_DEPTH_TEST);
	} else {
	    glDisable(GL_DEPTH_TEST);
	}
    }

    /* do surface shading */
    for (obj = port->cell->objects; obj; obj = obj->next) {

	if (obj->parent) continue;
	if (obj->invalid) continue;

	name = DrawObject(port, obj, name);
    }
}

/*
 *--------------------------------------------------------------
 *
 * DrawObject:
 *
 * recursivel draw an hierarchical object.
 *
 *--------------------------------------------------------------
 */

static int
DrawObject(port, obj, name)
     PortItem *port;
     Object *obj;
     GLuint name;
{
    int i, j, *index;
    Vertex *v;
    Surface *s;
    Object *child;
    Tcl_HashEntry *e;

    glPushMatrix();

#ifdef USE_DOUBLE
    glMultMatrixd(obj->modelT);
#else
    glMultMatrixf(obj->modelT);
#endif

    e = Tcl_FindHashEntry(&port->visibleHash, obj->label);
    if (obj->visible
	&& (((e == NULL) && !port->invertedVisibleMask) ||
	    ((e != NULL) && port->invertedVisibleMask))) {

	/* object shading model */

	switch(obj->shading) {
	  case SM_OBJECT_SHADING_FLAT:
	    glShadeModel(GL_FLAT);
	    break;

	  case SM_OBJECT_SHADING_SMOOTH:
	    glShadeModel(GL_SMOOTH);
	    break;
	}

	/* select front facing orientation */
	if (obj->ccw) {
	    glFrontFace(GL_CCW);
	} else {
	    glFrontFace(GL_CW);
	}

	/* front/back face culling */
	if (obj->cull) {
	    glCullFace(obj->cull);
	    glEnable(GL_CULL_FACE);
	} else {
	    glDisable(GL_CULL_FACE);
	}
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	/* polygon drawing mode */
	glPolygonMode(GL_FRONT, obj->frontmode);
	glPolygonMode(GL_BACK, obj->backmode);

	/* assign name for picking, if necessary */
	if (name != 0) {
	    obj->pickname = name;
	    glLoadName(name++);
	    glDisable(GL_LIGHTING);
	} else {

	    glEnable(GL_LIGHTING);

	    /* Object material */
	    glMaterialfv(GL_FRONT, GL_AMBIENT, obj->ambientRGBA);
	    glMaterialfv(GL_BACK, GL_AMBIENT, obj->bfambientRGBA);
	
	    glMaterialfv(GL_FRONT, GL_DIFFUSE, obj->diffuseRGBA);
	    glMaterialfv(GL_BACK, GL_DIFFUSE, obj->bfdiffuseRGBA);

	    glMaterialfv(GL_FRONT, GL_EMISSION, obj->emissionRGBA);
	    glMaterialfv(GL_BACK, GL_EMISSION, obj->bfemissionRGBA);
	    
	    glMaterialfv(GL_FRONT, GL_SPECULAR, obj->specularRGBA);
	    glMaterialfv(GL_BACK, GL_EMISSION, obj->bfspecularRGBA);
	    
	    glMaterialf(GL_FRONT, GL_SHININESS, (float) obj->specular.exponent);
	    glMaterialf(GL_BACK, GL_SHININESS, (float) obj->bfspecular.exponent);
	}

	if (obj->type == SM_TYPE_CYLINDER) {
	    gluQuadricNormals(obj->quadric, obj->quadNormals);
	    gluQuadricOrientation(obj->quadric, obj->orientation);
	    gluQuadricDrawStyle(obj->quadric, obj->drawStyle);
	    gluCylinder(obj->quadric, (GLdouble) obj->baseRadius,
			(GLdouble) obj->topRadius, (GLdouble) obj->height,
			(GLint) obj->slices, (GLint) obj->stacks);
	}
	else if (obj->type == SM_TYPE_DISK) {
	    gluQuadricNormals(obj->quadric, obj->quadNormals);
	    gluQuadricOrientation(obj->quadric, obj->orientation);
	    gluQuadricDrawStyle(obj->quadric, obj->drawStyle);
	    gluDisk(obj->quadric, (GLdouble) obj->innerRadius,
		    (GLdouble) obj->outerRadius, (GLint) obj->slices, (GLint) obj->loops);
	}
	else if (obj->type == SM_TYPE_SPHERE) {
	    gluQuadricNormals(obj->quadric, obj->quadNormals);
	    gluQuadricOrientation(obj->quadric, obj->orientation);
	    gluQuadricDrawStyle(obj->quadric, obj->drawStyle);
	    gluSphere(obj->quadric, (GLdouble) obj->radius,
		    (GLint) obj->slices, (GLint) obj->stacks);
	}
	else if (obj->type == SM_TYPE_POLYGON) {
	    for (i = 0, s = obj->s; i < obj->ns; i++, s++) {
		glBegin(GL_POLYGON);
		if (s->normal == SM_VERTEX_NORMAL_DEFAULT) {
		    if (obj->useAverage) {

			/* use per-vertex normal obtained from adjacent surface normals */
			for (j = 0, v = obj->v, index = s->index; j < s->vcnt; j++, index++) {

#ifdef USE_DOUBLE
			    glNormal3dv(v[*index].lvn.v);
			    glVertex3dv(v[*index].lv);
#else
			    glNormal3fv(v[*index].lvn.v);
			    glVertex3fv(v[*index].lv);
#endif
			}
		    } else {

			/* use surface normal */

#ifdef USE_DOUBLE
			glNormal3dv(s->ln);
#else
			glNormal3fv(s->ln);
#endif
			for (j = 0, v = obj->v, index = s->index; j < s->vcnt; j++, index++) {

#ifdef USE_DOUBLE
			    glVertex3dv(v[*index].lv);
#else
			    glVertex3fv(v[*index].lv);
#endif
			}
		    }
		} else {
		
		    /* use user-specified vertex normals */
		    for (j = 0, v = obj->v, index = s->index; j < s->vcnt; j++, index++) {

#ifdef USE_DOUBLE
			glNormal3dv(s->lvn[j].v);
			glVertex3dv(v[*index].lv);
#else
			glNormal3fv(s->lvn[j].v);
			glVertex3fv(v[*index].lv);
#endif
		    }
		}
		glEnd();
	    }
	}
    }

    /* recursively draw children with current object's model transform in effect */
    for (child = obj->child; child; child = child->sibling) {
	name = DrawObject(port, child, name);
    }
    glPopMatrix();

    return name;
}

/*
 *--------------------------------------------------------------
 *
 * SmCreateDrawable:
 *
 * create window/pixmap used for rendering by OpenGL/Mesa.
 *
 *--------------------------------------------------------------
 */


void
SmCreateDrawable(port)
     PortItem *port;
{
    Window win;
    Tk_Window tkwin;
    Display *dpy = port->dpy;

#ifdef MESA
    XWindowAttributes attr;
    unsigned long pixel;
#endif

    if (port->glxpmGC != None) {
	glXDestroyContext(dpy, port->glxpmGC);
	port->glxpmGC = None;
    }
    if (port->canvas) {

	/* if canvas item, use pixmap as off-screen buffer */
	if (port->pm != None) {
	    XFreePixmap(dpy, port->pm);
	    port->pm = None;
	}
	if (port->glxpm != None) {
	    glXDestroyGLXPixmap(dpy, port->glxpm);
	    port->glxpm = None;
	}
	tkwin = Tk_CanvasTkwin(port->canvas);

	port->pm = XCreatePixmap(dpy, Tk_WindowId(tkwin), port->w, port->h, port->visinfo->depth);
	port->glxpm = glXCreateGLXPixmap(dpy, port->visinfo, port->pm);
	port->glxpmGC = glXCreateContext(dpy, port->visinfo, NULL, False);


#ifdef MESA

    /* create a dummy window to make Mesa use the right window */
    /* size.    					       */

	pixel = BlackPixel(dpy, port->visinfo->screen);
	XGetWindowAttributes(dpy, Tk_WindowId(tkwin), &attr);
	win = XCreateSimpleWindow(dpy, Tk_WindowId(tkwin),
				  0, 0, port->w, port->h, 0, pixel, pixel);
	XSetWindowColormap(dpy, win, attr.colormap);
	glXMakeCurrent(dpy, win, port->glxpmGC);
#endif

	glXMakeCurrent(dpy, port->glxpm, port->glxpmGC);

#ifdef MESA
	XDestroyWindow(dpy, win);
#endif

    } else {
	port->glxpmGC = glXCreateContext(dpy, port->visinfo, NULL, True);
	glXMakeCurrent(dpy, Tk_WindowId(port->tkwin), port->glxpmGC);
    }

    glClearDepth(1.0);
    glEnable(GL_NORMALIZE);
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
}
